@@ -72,6 +72,7 @@ class Ha(object):
         self.recovering = False
         self._async_response = CriticalTask()
         self._crash_recovery_executed = False
+        self._crash_recovery_started = None
         self._start_timeout = None
         self._async_executor = AsyncExecutor(self.state_handler.cancellable, self.wakeup)
         self.watchdog = patroni.watchdog
@@ -156,7 +157,11 @@ class Ha(object):
                 last_operation = self.state_handler.last_operation()
             except Exception:
                 logger.exception('Exception when called state_handler.last_operation()')
-        ret = self.dcs.update_leader(last_operation, self._leader_access_is_restricted)
+        try:
+            ret = self.dcs.update_leader(last_operation, self._leader_access_is_restricted)
+        except Exception:
+            logger.exception('Unexpected exception raised from update_leader, please report it as a BUG')
+            ret = False
         self.set_is_leader(ret)
         if ret:
             self.watchdog.keepalive()
@@ -329,6 +334,7 @@ class Ha(object):
                 and not self._crash_recovery_executed and \
                 (self.cluster.is_unlocked() or self._rewind.can_rewind):
             self._crash_recovery_executed = True
+            self._crash_recovery_started = time.time()
             msg = 'doing crash recovery in a single user mode'
             return self._async_executor.try_run_async(msg, self._rewind.ensure_clean_shutdown) or msg
 
@@ -761,8 +767,10 @@ class Ha(object):
         if self.state_handler.is_starting():  # postgresql still starting up is unhealthy
             return False
 
-        if self.state_handler.is_leader():  # leader is always the healthiest
-            return True
+        if self.state_handler.is_leader():
+            # in pause leader is the healthiest only when no initialize or sysid matches with initialize!
+            return not self.is_paused() or not self.cluster.initialize\
+                    or self.state_handler.sysid == self.cluster.initialize
 
         if self.is_paused():
             return False
@@ -1165,6 +1173,14 @@ class Ha(object):
         Figure out what to do with the task AsyncExecutor is performing.
         """
         if self.has_lock() and self.update_lock():
+            if self._async_executor.scheduled_action == 'doing crash recovery in a single user mode':
+                time_left = self.patroni.config['master_start_timeout'] - (time.time() - self._crash_recovery_started)
+                if time_left <= 0 and self.is_failover_possible(self.cluster.members):
+                    logger.info("Demoting self because crash recovery is taking too long")
+                    self.state_handler.cancellable.cancel(True)
+                    self.demote('immediate')
+                    return 'terminated crash recovery because of startup timeout'
+
             return 'updated leader lock during ' + self._async_executor.scheduled_action
         elif not self.state_handler.bootstrapping:
             # Don't have lock, make sure we are not promoting or starting up a master in the background
@@ -1348,6 +1364,8 @@ class Ha(object):
                     self.release_leader_key_voluntarily()
                     return 'released leader key voluntarily as data dir empty and currently leader'
 
+                if self.is_paused():
+                    return 'running with empty data directory'
                 return self.bootstrap()  # new node
             else:
                 # check if we are allowed to join
@@ -1359,9 +1377,16 @@ class Ha(object):
 
                 if self.sysid_valid(self.cluster.initialize):
                     if self.cluster.initialize != data_sysid:
-                        logger.fatal("system ID mismatch, node %s belongs to a different cluster: %s != %s",
-                                     self.state_handler.name, self.cluster.initialize, data_sysid)
-                        sys.exit(1)
+                        if self.is_paused():
+                            logger.warning('system ID has changed while in paused mode. Patroni will exit when resuming'
+                                           ' unless system ID is reset: %s != %s', self.cluster.initialize, data_sysid)
+                            if self.has_lock():
+                                self.release_leader_key_voluntarily()
+                                return 'released leader key voluntarily due to the system ID mismatch'
+                        else:
+                            logger.fatal('system ID mismatch, node %s belongs to a different cluster: %s != %s',
+                                         self.state_handler.name, self.cluster.initialize, data_sysid)
+                            sys.exit(1)
                 elif self.cluster.is_unlocked() and not self.is_paused():
                     # "bootstrap", but data directory is not empty
                     if not self.state_handler.cb_called and self.state_handler.is_running() \
